"use strict"; // Assuming all the static methods from below array are side effect free evaluation
// except Math.random

function _slicedToArray(arr, i) { return _arrayWithHoles(arr) || _iterableToArrayLimit(arr, i) || _unsupportedIterableToArray(arr, i) || _nonIterableRest(); }

function _nonIterableRest() { throw new TypeError("Invalid attempt to destructure non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method."); }

function _iterableToArrayLimit(arr, i) { var _i = arr == null ? null : typeof Symbol !== "undefined" && arr[Symbol.iterator] || arr["@@iterator"]; if (_i == null) return; var _arr = []; var _n = true; var _d = false; var _s, _e; try { for (_i = _i.call(arr); !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"] != null) _i["return"](); } finally { if (_d) throw _e; } } return _arr; }

function _arrayWithHoles(arr) { if (Array.isArray(arr)) return arr; }

function _createForOfIteratorHelper(o, allowArrayLike) { var it = typeof Symbol !== "undefined" && o[Symbol.iterator] || o["@@iterator"]; if (!it) { if (Array.isArray(o) || (it = _unsupportedIterableToArray(o)) || allowArrayLike && o && typeof o.length === "number") { if (it) o = it; var i = 0; var F = function F() {}; return { s: F, n: function n() { if (i >= o.length) return { done: true }; return { done: false, value: o[i++] }; }, e: function e(_e2) { throw _e2; }, f: F }; } throw new TypeError("Invalid attempt to iterate non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method."); } var normalCompletion = true, didErr = false, err; return { s: function s() { it = it.call(o); }, n: function n() { var step = it.next(); normalCompletion = step.done; return step; }, e: function e(_e3) { didErr = true; err = _e3; }, f: function f() { try { if (!normalCompletion && it.return != null) it.return(); } finally { if (didErr) throw err; } } }; }

function _unsupportedIterableToArray(o, minLen) { if (!o) return; if (typeof o === "string") return _arrayLikeToArray(o, minLen); var n = Object.prototype.toString.call(o).slice(8, -1); if (n === "Object" && o.constructor) n = o.constructor.name; if (n === "Map" || n === "Set") return Array.from(o); if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen); }

function _arrayLikeToArray(arr, len) { if (len == null || len > arr.length) len = arr.length; for (var i = 0, arr2 = new Array(len); i < len; i++) arr2[i] = arr[i]; return arr2; }

const VALID_CALLEES = ["String", "Number", "Math"];
const INVALID_METHODS = ["random"];
const newIssueUrl = "https://github.com/babel/minify/issues/new";

module.exports = function ({
  types: t
}) {
  class BuiltInReplacer {
    constructor() {
      // map<expr_name, path[]>;
      this.pathsToUpdate = new Map();
    }

    getCollectVisitor() {
      const context = this;
      const collectVisitor = {
        AssignmentExpression(path) {
          const left = path.node.left; // Should bail and not run the plugin
          // when builtin is polyfilled

          if (t.isMemberExpression(left) && isBuiltInComputed(left)) {
            let parent = path;

            do {
              parent.stop();
            } while (parent = parent.parentPath);
          }
        },

        MemberExpression(path) {
          if (path.parentPath.isCallExpression()) {
            return;
          }

          const node = path.node;

          if (!isComputed(node) && isBuiltin(node) && !getFunctionParent(path).isProgram()) {
            const expName = memberToString(node);
            addToMap(context.pathsToUpdate, expName, path);
          }
        },

        CallExpression: {
          exit(path) {
            const callee = path.get("callee");

            if (!callee.isMemberExpression()) {
              return;
            }

            const node = callee.node; // computed property should not be optimized
            // Math[max]() -> Math.max()

            if (!isComputed(node) && isBuiltin(node) && !getFunctionParent(callee).isProgram()) {
              const expName = memberToString(node);
              addToMap(context.pathsToUpdate, expName, callee);
            }
          }

        }
      };
      return collectVisitor;
    }

    replace() {
      var _iterator = _createForOfIteratorHelper(this.pathsToUpdate),
          _step;

      try {
        for (_iterator.s(); !(_step = _iterator.n()).done;) {
          const _step$value = _slicedToArray(_step.value, 2),
                expName = _step$value[0],
                paths = _step$value[1];

          // transform only if there is more than 1 occurence
          if (paths.length <= 1) {
            continue;
          }

          const segmentsMap = getSegmentedSubPaths(paths);

          var _iterator2 = _createForOfIteratorHelper(segmentsMap),
              _step2;

          try {
            for (_iterator2.s(); !(_step2 = _iterator2.n()).done;) {
              const _step2$value = _slicedToArray(_step2.value, 2),
                    parent = _step2$value[0],
                    subpaths = _step2$value[1];

              if (subpaths.length <= 1) {
                continue;
              }

              const uniqueIdentifier = parent.scope.generateUidIdentifier(expName);
              const newNode = t.variableDeclaration("var", [t.variableDeclarator(uniqueIdentifier, subpaths[0].node)]);

              var _iterator3 = _createForOfIteratorHelper(subpaths),
                  _step3;

              try {
                for (_iterator3.s(); !(_step3 = _iterator3.n()).done;) {
                  const path = _step3.value;
                  path.replaceWith(t.clone(uniqueIdentifier));
                } // hoist the created var to the top of the function scope

              } catch (err) {
                _iterator3.e(err);
              } finally {
                _iterator3.f();
              }

              const target = parent.get("body");
              /**
               * Here, we validate a case where there is a local binding of
               * one of Math, String or Number. Here we have to get the
               * global Math instead of using the local one - so we do the
               * following transformation
               *
               * var _Mathmax = Math.max;
               *
               * to
               *
               * var _Mathmax = (0, eval)("this").Math.max;
               */

              var _iterator4 = _createForOfIteratorHelper(VALID_CALLEES),
                  _step4;

              try {
                for (_iterator4.s(); !(_step4 = _iterator4.n()).done;) {
                  const builtin = _step4.value;

                  if (target.scope.getBinding(builtin)) {
                    const prev = newNode.declarations[0].init;

                    if (!t.isMemberExpression(prev)) {
                      throw new Error(`minify-builtins expected a MemberExpression. ` + `Found ${prev.type}. ` + `Please report this at ${newIssueUrl}`);
                    }

                    if (!t.isMemberExpression(prev.object)) {
                      newNode.declarations[0].init = t.memberExpression(t.memberExpression(getGlobalThis(), prev.object), prev.property);
                    }
                  }
                }
              } catch (err) {
                _iterator4.e(err);
              } finally {
                _iterator4.f();
              }

              target.unshiftContainer("body", newNode);
            }
          } catch (err) {
            _iterator2.e(err);
          } finally {
            _iterator2.f();
          }
        }
      } catch (err) {
        _iterator.e(err);
      } finally {
        _iterator.f();
      }
    }

  }

  const builtInReplacer = new BuiltInReplacer();
  return {
    name: "minify-builtins",
    visitor: Object.assign({}, builtInReplacer.getCollectVisitor(), {
      Program: {
        exit() {
          builtInReplacer.replace();
        }

      }
    })
  };

  function memberToString(memberExprNode) {
    const object = memberExprNode.object,
          property = memberExprNode.property;
    let result = "";
    if (t.isIdentifier(object)) result += object.name;
    if (t.isMemberExpression(object)) result += memberToString(object);
    if (t.isIdentifier(property)) result += property.name;
    return result;
  }

  function isBuiltInComputed(memberExprNode) {
    const object = memberExprNode.object,
          computed = memberExprNode.computed;
    return computed && t.isIdentifier(object) && VALID_CALLEES.indexOf(object.name) >= 0;
  }

  function isBuiltin(memberExprNode) {
    const object = memberExprNode.object,
          property = memberExprNode.property;

    if (t.isIdentifier(object) && t.isIdentifier(property) && VALID_CALLEES.indexOf(object.name) >= 0 && INVALID_METHODS.indexOf(property.name) < 0) {
      return true;
    }

    return false;
  } // Creates a segmented map that contains the earliest common Ancestor
  // as the key and array of subpaths that are descendats of the LCA as value


  function getSegmentedSubPaths(paths) {
    let segments = new Map(); // Get earliest Path in tree where paths intersect

    paths[0].getDeepestCommonAncestorFrom(paths, (lastCommon, index, ancestries) => {
      // found the LCA
      if (!lastCommon.isProgram()) {
        let fnParent;

        if (lastCommon.isFunction() && t.isBlockStatement(lastCommon.node.body)) {
          segments.set(lastCommon, paths);
          return;
        } else if (!(fnParent = getFunctionParent(lastCommon)).isProgram() && t.isBlockStatement(fnParent.node.body)) {
          segments.set(fnParent, paths);
          return;
        }
      } // Deopt and construct segments otherwise


      var _iterator5 = _createForOfIteratorHelper(ancestries),
          _step5;

      try {
        for (_iterator5.s(); !(_step5 = _iterator5.n()).done;) {
          const ancestor = _step5.value;
          const fnPath = getChildFuncion(ancestor);

          if (fnPath === void 0) {
            continue;
          }

          const validDescendants = paths.filter(p => {
            return p.isDescendant(fnPath);
          });
          segments.set(fnPath, validDescendants);
        }
      } catch (err) {
        _iterator5.e(err);
      } finally {
        _iterator5.f();
      }
    });
    return segments;
  }

  function getChildFuncion(ancestors = []) {
    var _iterator6 = _createForOfIteratorHelper(ancestors),
        _step6;

    try {
      for (_iterator6.s(); !(_step6 = _iterator6.n()).done;) {
        const path = _step6.value;

        if (path.isFunction() && t.isBlockStatement(path.node.body)) {
          return path;
        }
      }
    } catch (err) {
      _iterator6.e(err);
    } finally {
      _iterator6.f();
    }
  }
  /**
   * returns
   *
   * (0, eval)("this")
   */


  function getGlobalThis() {
    return t.callExpression(t.sequenceExpression([t.valueToNode(0), t.identifier("eval")]), [t.valueToNode("this")]);
  }
};

function addToMap(map, key, value) {
  if (!map.has(key)) {
    map.set(key, []);
  }

  map.get(key).push(value);
}

function isComputed(node) {
  return node.computed;
}
/**
 * Babel-7 returns null if there is no function parent
 * and uses getProgramParent to get Program
 */


function getFunctionParent(path) {
  return (path.scope.getFunctionParent() || path.scope.getProgramParent()).path;
}